/*
 * (C) Copyright 2006 Graphics Cluster Rendering Project
 * (C) Copyright 2006 Faculty of Graphic Arts, University of Zagreb, Croatia
 *
 * This code is part of Graphics Rendering Cluster Project support, carried out
 * at Faculty of Graphic Arts, University of Zagreb, Croatia.
 *
 * This code is developed under GPL (GNU General Public License). It is
 * essential that this project remains under this license, as demanded by
 * license rights of code that was inherited to this project from other
 * GPL projects, as stated and credited elsewhere.
 *
 * Coding: Mirsad Todorovac, <mtodorov@grf.hr>
 *
 * 2006/Feb/06 Mon (Started in grc-client.cc)
 * 2006/Feb/07 Tue (spawned from grc-client.cc)
 *
 */

/* filecopy.cc -- what is needed for efficient transfer of files
 *                fro/to streams/files/descriptors
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <fcntl.h>
#include <limits.h>

#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "config.h"
#include "grc.h"

#define MAXDATASIZE 16384 // of course this will be dynamic in real use
#define PREFERRED_BUFSIZE (65536*4L)

int    _argc;
char **_argv;

int dispatch (char *model, int conn_sock) {

	int          fd;
	struct stat  stat_buf;
	void        *file_addr;

	printf ("entered dispatch()\n");

	if ((fd = open (model, O_RDONLY)) == -1) {
	    fprintf (stderr, "%s: Could not open model %s.\n", _argv[0], model);
	    perror ("open");
	    exit (105);
	}

	printf ("open pass\n");

	if (fstat (fd, &stat_buf) == -1) {
	    fprintf (stderr, "%s: ", _argv[0]);
	    perror ("fstat");
	    exit (106);
	}

	printf ("stat pass\n");

	file_addr = mmap (0, stat_buf.st_size, PROT_READ, MAP_SHARED
#ifdef MAP_POPULATE
			|MAP_POPULATE
#endif
			, fd, 0);

	printf ("mmap pass\n");
	
	if (file_addr != MAP_FAILED) {
	    ssize_t remain          = stat_buf.st_size;
	    ssize_t written         = 0;
	    char *sendbuf           = (char *)file_addr;
	    int   semaphore = 0;
	   
	    close (fd); 
	    fprintf (stderr, "conn %d: written=%d remain=%d\n",
				conn_sock,
				written, remain);
	    while (remain > 0) {
		printf ("entering write() call\n");
		alarm (60);
		written = write (conn_sock, sendbuf, remain);
		if (written == -1) {
		    // krk
		    fprintf (stderr, "%s: ", _argv[0]);
		    perror ("write mmaped region to conn_sock failed");
		    if (errno == EINTR || errno == EAGAIN) {
			sleep (1);
			errno = 0;
			continue;
		    } else
		        return written;
		} else if (written == 0) {
		    perror ("write mmaped region to conn_sock 0 bytes");
		    if (semaphore == 1) {
			sleep (1);
		    } else {
		        semaphore = 1;
	            }
		} else if (written > remain) {
		    fprintf (stderr, "%s: Internal program error 0xD0000001\n");
		    exit (108);
		} else {
		    sendbuf += written;
		    remain -= written;
		    semaphore = 0;
	            fprintf (stderr, "conn %d: written=%d remain=%d\n",
				conn_sock, written, remain);
		}
		printf ("out of write() loop\n");
	    }
	    return written;

	} else { // MAP_FAILED
	    char     *rdsn_buffer = NULL;
	    ssize_t   totalread = 0, transferred = 0;  
	    // int       bufsize = PREFERRED_BUFSIZE;
	    char      buf [65536*4];
	    ssize_t   bufsize = sizeof (buf);
	    ssize_t   actual = 0, actual_wr = 0, total_xfer = 0, readsize = 0,
		      filesize = 0, total_wr = 0;

	    perror ("Non fatal: mmap failed"); // we want to inform people
/*
	    // not fatal, try the failsafe read/write copying of file to socket
	    rdsn_buffer = calloc (bufsize, 1);
	    while (rdsn_buffer == NULL && bufsize > 1) {
		bufsize = bufsize / 2;             // we try ot be really
	        rdsn_buffer = calloc (bufsize, 1); // robust and persistant
            }
	    if (rdsn_buffer == NULL) {
		fprintf (stderr, "%s: ", _argv[0]);
		if (errno == ENOMEM) {
		    perror ("malloc");
		} else {
		    fprintf (stderr, "malloc: Not enough memory.\n");
		}
		exit (107);
            }

	    // OK; we survived allocating the buffer, now try the transfer

	    while ((actual = read (fd, rdsn_buf, bufsize)) > 0) {
		if (actual == -1) {
		    fprintf (stderr, "%s: ", _argv[0]);
		    perror ("model read");
		    exit (107);
		}
		if (actual == 0 && totalread < stat_buf->st_size) {
		    fprintf (stderr, "%s: %s: File shrank?\n", _argv[0], model);
		    exit (108);
		}
		if (actual == 0 && totalread > stat_buf->st_size) {
		    fprintf (stderr, "%s: %s: File grown?\n", _argv[0], model);
		    exit (109);
		}
		totalread += actual;
		actual_wr = write (conn_sock, rdsn_buf, totalread);
		if (actual_wr == -1) {
		    fprintf (stderr, "%s: ", _argv[0]);
		    perror ("render srv conn write);
		    return actual_wr;
		}
	    }

*/
	    filesize = stat_buf.st_size;
	    readsize = MIN (filesize, bufsize);
	    fprintf (stderr, "total = %d, actrd = %d, readsz = %d actwr=%d tot_wr=%d\n", total_xfer, actual, readsize, actual_wr, total_wr);
	
	    while (total_xfer < filesize &&
	           (actual = read (fd, buf, readsize)) > 0) {
	        actual_wr = write (conn_sock, buf, actual); // reinforce later!
		if (actual_wr == -1) {
			perror ("write on socket");
			//if (errno == EINTR || errno == EAGAIN) {
			//    errno = 0;
			//    continue; // retry
			//}
			return -1;
		}
		else if (actual_wr == 0)
			fprintf (stderr, "0 write to connection\n");
	        total_xfer += actual;
		total_wr += actual_wr;
	        readsize = MIN (filesize - total_xfer, bufsize);
	        fprintf (stderr, "total = %d, actrd = %d, readsz = %d actwr=%d tot_wr=%d\n", total_xfer, actual, readsize, actual_wr, total_wr);
	    }
	    fprintf (stderr, "total = %d, actrd = %d, readsz = %d actwr=%d tot_wr=%d\n", total_xfer, actual, readsize, actual_wr, total_wr);
	    close (fd);
        }
}
		
int receive (int conn_fd, char *filename)
{
	int     fd;
	char    buf[65536*4];
	ssize_t bufsize = sizeof (buf);
	ssize_t actual;

	fd = open (filename, O_RDWR|O_CREAT|O_TRUNC, 0666);
	if (fd == -1) {
	    fprintf (stderr, "%s: ", _argv[0]);
	    perror ("open storage");
	    if (send (conn_fd, "500 Cannot open file.\n", 22, 0) == -1)
		perror ("send");
	    close (conn_fd);
	    exit (108);
	}

	// this needs some reinforcement, but make it run
	// ASAP for now!
	while ((actual = read (conn_fd, buf, bufsize)) > 0) {
	    write (fd, buf, actual);
	}
}

int receive (int conn_fd, char *filename, ssize_t filesize)
{
	int     fd;
	char    buf[65536*4];
	ssize_t bufsize = sizeof (buf);
	ssize_t actual = 0, actual_wr = 0, total_xfer = 0, readsize = 0,
		total_wr = 0, act_fsz = 0;

	act_fsz = filesize;

	fd = open (filename, O_RDWR|O_CREAT|O_TRUNC, 0666);
	if (fd == -1) {
	    fprintf (stderr, "%s: ", _argv[0]);
	    perror ("open storage");
	    if (send (conn_fd, "500 Cannot open file.\n", 22, 0) == -1)
		perror ("send");
	    close (conn_fd);
	    exit (108);
	}

	// this needs some reinforcement, but make it run
	// ASAP for now!
	
	readsize = MIN (act_fsz, bufsize);
	    fprintf (stderr,
	"total = %d, actual = %d, readsz = %d actwr=%d tot_wr=%d flesize=%d\n",
		    total_xfer, actual, readsize, actual_wr, total_wr, act_fsz);
	
	while (total_xfer < act_fsz)
	{
	    printf ("read %ld/%ld:", total_xfer, act_fsz); fflush (stdout);
	    if ((actual = read (conn_fd, buf, readsize)) == -1) {
	        perror ("read");
		//if (errno == EAGAIN || errno == EINTR) {
		//    sleep (1);
		//    continue;
		//} else
		    return -1;
	    }
		
	    if (actual == 0) {
		fprintf (stderr, "Zero read, " __FILE__ ": %d\n", __LINE__);
		break;
	    }
	    if ((actual_wr = write (fd, buf, actual)) == -1)
		    perror ("write");
	    total_xfer += actual;
	    total_wr += actual_wr;
	    readsize = MIN (act_fsz - total_xfer, bufsize);
	    fprintf (stderr, "total = %d, actual = %d, readsz = %d actwr=%d tot_wr=%d\n",
			    total_xfer, actual, readsize, actual_wr, total_wr);
	    if (total_xfer == act_fsz) printf ("equal\n");
	}
	fprintf (stderr,
		"fin total = %d, actual = %d, readsz = %d actwr=%d tot_wr=%d\n",
		    total_xfer, actual, readsize, actual_wr, total_wr);
	close (fd);
	return total_wr;
}

